//# ComponentType.h: Enumerators common to the Components Module
//# Copyright (C) 1997,1998,1999,2002
//# Associated Universities, Inc. Washington DC, USA.
//#
//# This library is free software; you can redistribute it and/or modify it
//# under the terms of the GNU Library General Public License as published by
//# the Free Software Foundation; either version 2 of the License, or (at your
//# option) any later version.
//#
//# This library is distributed in the hope that it will be useful, but WITHOUT
//# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
//# License for more details.
//#
//# You should have received a copy of the GNU Library General Public License
//# along with this library; if not, write to the Free Software Foundation,
//# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
//#
//# Correspondence concerning AIPS++ should be addressed as follows:
//#        Internet email: aips2-request@nrao.edu.
//#        Postal address: AIPS++ Project Office
//#                        National Radio Astronomy Observatory
//#                        520 Edgemont Road
//#                        Charlottesville, VA 22903-2475 USA
//#
//#
//# $Id: ComponentType.h 21130 2011-10-18 07:39:05Z gervandiepen $

#ifndef COMPONENTS_COMPONENTTYPE_H
#define COMPONENTS_COMPONENTTYPE_H

#include <casacore/casa/aips.h>

//# Forward Declaration
namespace casacore{

class String;
}

namespace casac {
      class componentlist;
}

namespace casa {


//# Forward Declaration
class SkyCompRep;
class ComponentShape;
class SpectralModel;

// <summary>Enumerators common to the ComponentsModels module</summary>

// <use visibility=export>

// <reviewed reviewer="" date="yyyy/mm/dd" tests="tComponentType" demos="">
// </reviewed>

// <etymology>
// This class contains the type definitions in the ComponentModels module 
// </etymology>

// <synopsis> 
// This class does nothing.  It is merely a container for the enumerations used
// by the ComponentModels module.  These enumerations define the standard
// component types. It also contains:
// <ul>
// <li> static functions which convert between these enumerators and strings.
// <li> static functions which construct the appropriate derived object given
//      an enumerator. Note that these functions are only accessible by the
//      friend classes <linkto class=SkyCompRep>SkyCompRep</linkto> and
//      <linkto class=ComponentList>ComponentList</linkto> and should be
//      considered an implementation detail.
// </ul>
// The first element in the enumerator must be represented by zero and every
// enumerator must contain as the second last and last elements an UNKNOWN_*
// and NUMBER_* element.  
// </synopsis>

// <example>
// See the documentation for the ComponentModels classes for examples on the
// use of these enumerator and functions. For example the 
// <linkto class=ComponentShape>ComponentShape</linkto> class.
// </example>

// <todo asof="1999/11/15">
// <li> Nothing I can think of.
// </todo>
//

class ComponentType {
public:
  // Declare which classes access the private functions which convert
  // enumerators to objects. The private construct functions can only be
  // accessed by functions in the friend classes specified below.
  friend class SkyCompRep;
  friend class ::casac::componentlist;

  // The shapes of all the components
  enum Shape {
    // A simple point component
    POINT = 0,
    // A elliptical Gaussian component
    GAUSSIAN,
    // A elliptical disk component
    DISK,
    // A elliptical disk component with limb-darkening effect
    LDISK,
    // An unknown Component
    UNKNOWN_SHAPE,
    // The number of types in this enumerator
    NUMBER_SHAPES
  };
  // The ways the Flux polarisation can be represented
  enum Polarisation {
    // The four casacore::Stokes parameters, ie I,Q,U,V 
    STOKES = 0,
    // Linear polarisation feeds ie., XX,XY,YX,YY with zero parrallactic angle
    LINEAR,
    // Circular polarisation feeds ie., RR, RL, LR, LL
    CIRCULAR,
    // An unknown Component
    UNKNOWN_POLARISATION,
    // The number of types in this enumerator
    NUMBER_POLARISATIONS
  };
  // The different functional forms for the spectral variation.
  enum SpectralShape {
    // The flux is constant at different frequencies
    CONSTANT_SPECTRUM = 0,
    // The flux varies as <src>I = I_0 * (nu / nu_0)^alpha</src>
    SPECTRAL_INDEX,
    //Tabular values interpolated in between
    TABULAR_SPECTRUM,
    // Power Logarithmic Polynomial
    PLP,
    // an unknown spectral type
    UNKNOWN_SPECTRAL_SHAPE,
    // The number of spectral types in this enumerator
    NUMBER_SPECTRAL_SHAPES
  };
  // Convert the Shape enumerator to a string
  static casacore::String name(ComponentType::Shape shapeEnum);
  // Convert a given casacore::String to a Shape enumerator (min match active)
  static ComponentType::Shape shape(const casacore::String& shapeName);

  // Convert the Polarisation enumerator to a string
  static casacore::String name(ComponentType::Polarisation fluxEnum);
  // Convert a given casacore::String to a Polarisation enumerator
  static ComponentType::Polarisation polarisation(const casacore::String& 
						  polarisationName);
  // Convert the SpectralShape enumerator to a string
  static casacore::String name(ComponentType::SpectralShape spectralEnum);
  // Convert a given casacore::String to a SpectralShape enumerator
  static ComponentType::SpectralShape spectralShape(const casacore::String&
						    spectralName);
private:

  // Convert the Shape enumerator to a shape object (upcast to the base
  // object). Returns a null pointer if the object could not be
  // constructed. This will occur is the enumerator is UNKNOWN_SHAPE or
  // NUMBER_SHAPES or there is insufficient memory. The caller of this function
  // is responsible for deleting the pointer. 
  static ComponentShape* construct(ComponentType::Shape shapeEnum);

  // Convert the SpectralShape enumerator to a spectral model object (upcast to
  // the base object). Returns a null pointer if the object could not be
  // constructed. This will occur is the enumerator is UNKNOWN_SPECTRAL_SHAPE
  // or NUMBER_SPECTRAL_SHAPES or there is insufficient memory. The caller of
  // this function is responsible for deleting the pointer. 
  static SpectralModel* construct(ComponentType::SpectralShape spectralEnum);
};

} //# NAMESPACE CASA - END

#endif
